/* -*- Mode: C; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*  arclayer is a filter which takes a layer and bends it in an arc
 *  specified by a circle of a specified radius.
 *  It can arc along the top or bottom of the circle.
 *        Copyright (C) 2002  Akkana Peck.
 *
 *  To build: gimptool --build arclayer.c
 *  To install: gimptool --install-strip arclayer.c
 */

#ifdef IN_GIMP_TREE
#include "config.h"
#include "libgimp/stdplugins-intl.h"
#else
/* gimptool doesn't understand these: */
#define _(s)   s
#define N_(s)  s
#define INIT_I18N() 
#define INIT_I18N_UI() 
#endif

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/*
#ifdef __GNUC__
#warning GTK_DISABLE_DEPRECATED
#endif
#undef GTK_DISABLE_DEPRECATED
*/

#include <gtk/gtk.h>

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>

/* Backwards compatibility defines for gimp 1.2: */
#ifndef G_N_ELEMENTS
#define GIMP_1_2 1
#define MAYBE_CONST 
#define GimpRunMode gint32
#define drawable_id id
#define G_CALLBACK GTK_SIGNAL_FUNC
#define G_OBJECT GTK_OBJECT
#define g_signal_connect gtk_signal_connect
#define gtk_check_button_new_with_mnemonic gtk_check_button_new_with_label

#define GTK_STOCK_CANCEL _("Cancel")
#define GTK_STOCK_OK _("Ok")

#define G_N_ELEMENTS(a) ((sizeof a) / (sizeof (*(a))))

#else /* G_N_ELEMENTS, gimp 2.0 */
#define MAYBE_CONST const
#endif /* G_N_ELEMENTS, gimp 1.2 vs. 2.0 */

#define  SCALE_WIDTH     150

/*  The logo  */
#include "arclayer_icon.h"

typedef struct
{
  gint       run;
} ArclayerInterface;

typedef struct
{
    gint radius;
    gboolean top;
} ArclayerVals;

static ArclayerVals avals =
{
    400,      /* radius */
    TRUE      /* top */
};

/* Declare local functions.
 */
static void      query  (void);
static void      run    (MAYBE_CONST gchar        *name,
                                     gint          nparams,
                         MAYBE_CONST GimpParam    *param,
                               gint         *nreturn_vals,
                               GimpParam   **return_vals);
static gint32    make_arclayer (gint32 image_id, GimpDrawable *drawable,
                                gint32 radius, gboolean top);

/*  user interface functions  */
static gint      arclayer_dialog      (void);
static void      arclayer_ok_callback (GtkWidget *widget,
#ifndef GIMP_1_2
                                       gint response_id,
#endif
                                       gpointer   data);

static ArclayerInterface aint =
{
  FALSE,        /* run */
};

GimpPlugInInfo PLUG_IN_INFO =
{
  NULL,  /* init_proc  */
  NULL,  /* quit_proc  */
  query, /* query_proc */
  run,   /* run_proc   */
};

MAIN ()

static void
query (void)
{
  static GimpParamDef args[] =
  {
    { GIMP_PDB_INT32,    "run_mode", "Interactive, non-interactive" },
    { GIMP_PDB_IMAGE,    "image", "Input image" },
    { GIMP_PDB_DRAWABLE, "drawable", "Input drawable" },
    { GIMP_PDB_INT32,    "radius", "The radius of the arc" },
    { GIMP_PDB_INT32,    "top", "Draw along the top of the arc?" }
  };
  static GimpParamDef return_vals[] =
  {
    { GIMP_PDB_LAYER, "Arc Layer", "The new layer." }
  }; 

#ifdef IN_GIMP_TREE
  INIT_I18N();
#endif

  gimp_install_procedure ("plug_in_arclayer",
                          "Bend a layer along a circular arc",
                          "Help not yet written for this plug-in",
                          "Akkana Peck",
                          "Akkana Peck",
                          "2002",
                          N_("<Image>/Filters/Distorts/Arc Layer..."),
                          "RGB*, GRAY*, INDEXED*",
                          GIMP_PLUGIN,
                          G_N_ELEMENTS (args), G_N_ELEMENTS (return_vals),
                          args, return_vals);
}

static void
run (MAYBE_CONST gchar      *name,
                 gint        nparams,
     MAYBE_CONST GimpParam  *param,
                 gint       *nreturn_vals,
                 GimpParam **return_vals)
{
  static GimpParam values[2];
  GimpRunMode  run_mode;
  GimpPDBStatusType status = GIMP_PDB_SUCCESS;
  GimpDrawable *drawable;
  gint32 image_id;

  run_mode = param[0].data.d_int32;

  *nreturn_vals = 2;
  *return_vals  = values;
  values[0].type          = GIMP_PDB_STATUS;
  values[0].data.d_status = status;
  values[1].type = GIMP_PDB_LAYER;
  values[1].data.d_layer = -1;

  switch (run_mode)
  {
    case GIMP_RUN_INTERACTIVE:
      INIT_I18N_UI();
      /*  Possibly retrieve data  */
      gimp_get_data ("plug_in_arclayer", &avals);

      /*  First acquire information with a dialog  */
      if (! arclayer_dialog ())
	return;
      break;

    case GIMP_RUN_NONINTERACTIVE:
      INIT_I18N();
      /*  Make sure all the arguments are there!  */
      if (nparams != 15)
        status = GIMP_PDB_CALLING_ERROR;
      if (status == GIMP_PDB_SUCCESS)
      {
        avals.radius = param[3].data.d_int32;
        avals.top = (param[4].data.d_int32) ? TRUE : FALSE;
      }
      break;

    case GIMP_RUN_WITH_LAST_VALS:
      /*  Possibly retrieve data  */
      gimp_get_data ("plug_in_arclayer", &avals);
      break;

    default:
      break;
    }

  /*  Get the active drawable  */
  image_id = param[1].data.d_image;
  drawable = gimp_drawable_get (param[2].data.d_drawable);

  /*  Do it!  */
  if (status == GIMP_PDB_SUCCESS)
  {
    /*  run the effect  */
    values[1].data.d_layer = make_arclayer (image_id, drawable,
                                            avals.radius, avals.top);

    /*  If the run mode is interactive, flush the displays  */
    if (run_mode != GIMP_RUN_NONINTERACTIVE)
      gimp_displays_flush ();

    /*  Store avals data  */
    if (run_mode == GIMP_RUN_INTERACTIVE)
      gimp_set_data ("plug_in_arclayer", &avals, sizeof (ArclayerVals));
  }
  else if (status == GIMP_PDB_SUCCESS)
  {
    /* gimp_message ("mosaic: cannot operate on indexed color images"); */
    status = GIMP_PDB_EXECUTION_ERROR;
  }

  values[0].data.d_status = status;

  gimp_drawable_detach (drawable);
}

/* p_clear_drawable stolen from curve_bend */
void
p_clear_drawable (GimpDrawable *drawable)
{
   GimpPixelRgn  pixel_rgn;
   gpointer	pr;

#if 0
   gimp_pixel_rgn_init (&pixel_rgn, drawable,
                        0, 0, drawable->width, drawable->height,
                        TRUE,  /* dirty */
                        FALSE  /* shadow */
     );
#endif

   /* clear the drawable with 0 Bytes (black full-transparent pixels) */
   for (pr = gimp_pixel_rgns_register (1, &pixel_rgn);
        pr != NULL; pr = gimp_pixel_rgns_process (pr))
   {
     gint l_row;
     guchar *l_ptr = pixel_rgn.data;
     for ( l_row = 0; l_row < pixel_rgn.h; l_row++ )
     {
       memset(l_ptr, 0, pixel_rgn.w * drawable->bpp);
       l_ptr += pixel_rgn.rowstride;
     }
   }
}

/* Map coordinates, which we'll use in making the arc layer */
static void mapXYtoArcXY(int xx, int yy, int radius, int top, double theta2,
                         int old_h, int new_w, int new_h,
                         int* newx, int* newy)
{
  double phi = theta2 - (double)xx/radius;  /* current angle from center */
  int r;

  /* (xx, yy) is the position in the old layer;
   * map this to (newx, newy) in the new layer.
   */
  if (top)
  {
    r = radius - yy;
    *newy = (int)(radius - r * cos(phi));
  }
  else
  {
    r = radius - old_h + yy;
    *newy = new_h + (int)(r * cos(phi)) - radius;
  }
  if (*newy > new_h)
    *newy = new_h-1;
  else if (*newy < 0)
    *newy = 0;

  *newx = (int)(new_w/2 - r * sin(phi));

  if (*newx >= new_w)
    *newx = new_w-1;
  if (*newy >= new_h)
    *newy = new_h-1;
}

/* make_arclayer returns the layer ID of the newly created layer */
static gint32
make_arclayer (gint32 image_id, GimpDrawable *old_layer,
               gint32 radius, gboolean top)
{
  GimpPixelRgn  src_rgn;
  GimpPixelRgn  dest_rgn;
  guchar* dest = 0;
  GimpDrawable* arc_layer;
#ifdef GIMP_1_2
  gchar* old_name = gimp_layer_get_name(old_layer->drawable_id);
#else /* GIMP_1_2 */
  gchar* old_name = gimp_drawable_get_name(old_layer->drawable_id);
#endif /* GIMP_1_2 */
  gchar* new_name = g_malloc(strlen(old_name) + 5);
  double theta2 = .5 * old_layer->width / radius;
  gint new_width = (int)(2 * radius * sin(theta2));
  gint new_height = (int)(old_layer->height * cos(theta2)
                          + .5 * radius * sin(theta2) * tan(theta2));
  gint32 x1, y1;
  gint32 nlayers;
  gint32 stackpos;
  gint32* image_layers;
  gpointer pr;
  guchar* src;
  gint new_layer_id;
  double progress, max_progress;
  int memsize;
  int x0dest, y0dest, wdest, hdest;  /* limits of the tile when mapped */

#ifdef GIMP_1_2
  gimp_undo_push_group_start (image_id);
#else /* GIMP_1_2 */
  gimp_image_undo_group_start (image_id);
#endif /* GIMP_1_2 */

  sprintf(new_name, "%s arc", old_name);

  gimp_progress_init ( _("Bending arc..."));
  /*
  printf("Old layer: %d x %d, new layer: %d x %d\n",
         old_layer->width, old_layer->height, new_width, new_height);
   */

  image_layers = gimp_image_get_layers (image_id, &nlayers);
  stackpos = 0;
  while (stackpos < nlayers &&
         image_layers[stackpos] != old_layer->drawable_id)
    ++stackpos;

  /* Make the new layer the same color depth as the source: */
  new_layer_id = gimp_layer_new (image_id, new_name,
                                 new_width, new_height,
                                 gimp_drawable_type(old_layer->drawable_id),
                                 100, GIMP_NORMAL_MODE);
  gimp_image_add_layer (image_id, new_layer_id, stackpos+1);
  arc_layer = gimp_drawable_get (new_layer_id);
  /* printf("New drawable is %p, id %d\n", arc_layer, new_layer_id); */

  /* Various attempts to clear the layer first: */
  /*
  p_clear_drawable(arc_layer);
  gimp_selection_all(new_layer_id);
  gimp_edit_clear(new_layer_id);
  */

  max_progress = old_layer->width * old_layer->height;
  progress = 0;

  gimp_pixel_rgn_init (&src_rgn, old_layer,
                       0, 0, old_layer->width, old_layer->height,
                       FALSE, FALSE);
  gimp_pixel_rgn_init (&dest_rgn, arc_layer,
                       0, 0, new_width, new_height,
                       TRUE, TRUE);

  /* Decide whether to allocate the whole buffer, or go pixel by pixel,
   * depending on how big the new region needs to be.
   * This should probably check the tile cache size pref, or something.
   *
   * SMALL_TILES is an optimization -- it should allocate a chunk size
   * just big enough to handle everything that's mapped from this pr.
   * But it doesn't work and often crashes mysteriously,
   * so it's turned off for now.
   */
#ifndef SMALL_TILES
  memsize = new_width * new_height * dest_rgn.bpp;
  if (memsize < 130000000)
    dest = (guchar *)g_malloc(memsize);
  if (dest)
    bzero(dest, new_width * new_height * dest_rgn.bpp);
  else
    p_clear_drawable(arc_layer);
  x0dest = y0dest = 0;
  wdest = new_width;
  hdest = new_height;
#ifdef NOISY
  if (dest)
    printf("Allocating buffer\n");
  else
    printf("Going pixel by pixel\n");
#endif
#endif /* SMALL_TILES */

  /* Loop is from http://www.home.unix-ag.org/simon/gimp/guadec2002/gimp-plugin/html/efficientaccess.html
   * But if we register both regions it crashes, in gimp_pixel_rgns_process,
   * presumably because the src and dst regions are different sizes.
   */
  for (pr = gimp_pixel_rgns_register (1, &src_rgn);
       pr != NULL;
       pr = gimp_pixel_rgns_process (pr))
  {
#ifdef SMALL_TILES
    int x2, y2, x3, y3;                /* used for calculating the limits */
#endif /* SMALL_TILES */

    src = src_rgn.data;

#ifdef SMALL_TILES
# ifdef NOISY
    printf("\nChunk: (%d, %d) %d x %d\n", src_rgn.x, src_rgn.y,
           src_rgn.w, src_rgn.h);
# endif

    /* Allocate the dest region corresponding to this tile.
     * It will be bigger than the old tile, due to the arc
     * (so unfortunately we can't use the layer's natural tiling).
     */
    /* First, we need to map all four corners: */
    mapXYtoArcXY(src_rgn.x, src_rgn.y,
                 radius, top, theta2,
                 old_layer->height, new_width, new_height,
                 &x0dest, &y0dest);
    mapXYtoArcXY(src_rgn.x, src_rgn.y + src_rgn.h,
                 radius, top, theta2,
                 old_layer->height, new_width, new_height,
                 &x1, &y1);
    mapXYtoArcXY(src_rgn.x + src_rgn.w, src_rgn.y,
                 radius, top, theta2,
                 old_layer->height, new_width, new_height,
                 &x2, &y2);
    mapXYtoArcXY(src_rgn.x + src_rgn.w, src_rgn.y + src_rgn.h,
                 radius, top, theta2,
                 old_layer->height, new_width, new_height,
                 &x3, &y3);
    printf("Endpoints: (%d, %d), (%d, %d), (%d, %d),  (%d, %d)\n",
           x0dest, y0dest, x1, y1, x2, y2, x3, y3);

    /* swap if needed */
    x0dest = MIN(x0dest, x1);
    y0dest = MIN(y0dest, y2);
    x1 = MAX(x2, x3);
    y1 = MAX(y1, y3);

    printf("Region corners: (%d, %d) (%d, %d)\n", x0dest, y0dest, x1, y1);

    /* Make sure it extends far enough vertically.
     * If we crossed the midpoint, then neither endpoint is really
     * the vertical maximum, so fiddle that somehow.
     */
    x3 = src_rgn.x + src_rgn.h/2;  /* midpoint */
    if (x0dest < x3 && x1 > x3) {
      wdest = x1 - x0dest;
      hdest = y1 - y0dest;
      if (top)
        y0dest -= hdest/2;
      else
        y1 += hdest/2;
      hdest = y1 - y0dest + 1;
    }

    /* Sanity check of endpoints */
    x0dest = MAX(x0dest, 0);
    x1 = MIN(x1, new_width);
    y0dest = MAX(y0dest, 0);
    y1 = MIN(y1, new_height);

    /* Now we have the endpoints.  Get width and height. */
    wdest = x1 - x0dest;
    hdest = y1 - y0dest;

    printf("New region: (%d, %d) %d x %d\n", x0dest, y0dest, wdest, hdest);
    memsize = wdest * hdest * dest_rgn.bpp;
    printf("Trying to allocate %d bytes = %d * %d * %d\n",
           memsize, wdest, hdest, dest_rgn.bpp);
    dest = (guchar *)g_malloc(memsize);
    if (dest) {
      if (src_rgn.x == 0 && src_rgn.y == 0)
        bzero(dest, memsize);
      else
        gimp_pixel_rgn_get_rect(&dest_rgn, dest, x0dest, y0dest, wdest, hdest);
    }
    else printf("Couldn't allocate\n");
#endif /* SMALL_TILES */

    for (y1 = 0; y1 < src_rgn.h; y1++)
    {
      int yy = src_rgn.y + y1;
      guchar* s = src + y1 *  src_rgn.rowstride;
      int i;

      for (x1 = 0; x1 < src_rgn.w; x1++)
      {
        int xx = src_rgn.x + x1;
        int newx, newy;
        guchar* d = 0;
        guchar buf[dest_rgn.bpp];

        mapXYtoArcXY(xx, yy, radius, top, theta2,
                     old_layer->height, new_width, new_height,
                     &newx, &newy);

        if (dest) {
          int offset = ((newy - y0dest) * wdest
                        + (newx - x0dest)) * dest_rgn.bpp;
          if (offset > memsize) {
            printf("Yikes!  %d > %d\n", offset, memsize);
            printf("  (%d, %d) -> (%d, %d) in (%d, %d) %d x %d\n",
                   xx, yy, newx, newy, x0dest, y0dest, wdest, hdest);
            printf("  r=%d, theta2=%f, %d, %d, %d\n",
                   radius, (float)theta2,
                   old_layer->height, new_width, new_height);
            continue;
          }
          if (offset < 0) {
            printf("Yikes!  %d < 0\n", offset);
            continue;
          }
          d = dest + offset;
          /*printf("%d (%d - %d, %d - %d) ", d - dest, newx, x0dest, newy, y0dest);*/
          for (i=0; i<src_rgn.bpp; ++i)
            d[i] = s[i];
        }
        else {
          for (i=0; i<src_rgn.bpp; ++i)
            buf[i] = s[i];
          gimp_pixel_rgn_set_pixel(&dest_rgn, buf, newx, newy);
        }
        
        /* Here is a fast way to fill in holes:
         * Write to one pixel below the new location.
         * If that's a valid location, then later
         * another set of coordinates will map to it,
         * and overwrite this value; if not, then it
         * would have been a hole, and this fills it.
         */
        if (newy < new_height-2) {
          if (dest) {
            d[i + dest_rgn.w*dest_rgn.bpp] = s[i];
            d[i + 2*dest_rgn.w*dest_rgn.bpp] = s[i];
          }
          else {
            gimp_pixel_rgn_set_pixel(&dest_rgn, buf, newx, newy+1);
            gimp_pixel_rgn_set_pixel(&dest_rgn, buf, newx, newy+2);
          }
        }

        s += src_rgn.bpp;
      }
      progress += src_rgn.w;
#ifdef VERY_NOISY
      printf("progress: %f\n", progress / max_progress);
#endif
      gimp_progress_update (progress / max_progress);
    }

#ifdef SMALL_TILES
    /* Done with pr -- update the dest pixel region */
    if (dest) {
      printf("Setting in %p, id %d\n", arc_layer, new_layer_id);
      printf("Setting rect: (%d, %d), %d x %d\n", x0dest, y0dest, wdest,hdest);
      gimp_pixel_rgn_set_rect(&dest_rgn, dest, x0dest, y0dest, wdest, hdest);
      printf("freeing\n");
      g_free(dest);
      dest = 0;
#if 0
      gimp_drawable_flush (arc_layer);
      gimp_drawable_merge_shadow (new_layer_id, TRUE);
      gimp_drawable_update (new_layer_id, x0dest, y0dest, wdest, hdest);
#endif
    }
#endif /* SMALL_TILES */
  }

  gimp_progress_update (1.0);

  /*  update the drawable  */
  if (dest)
    gimp_pixel_rgn_set_rect(&dest_rgn, dest, 0, 0, new_width, new_height);

  gimp_image_remove_layer(image_id, old_layer->drawable_id);

  gimp_drawable_flush (arc_layer);
  gimp_drawable_merge_shadow (new_layer_id, TRUE);
  gimp_drawable_update (new_layer_id, 0, 0, new_width, new_height);

  gimp_displays_flush();
  gimp_drawable_detach(arc_layer);

  /* Make the center of the new layer contiguous with the center of the old: */
  gimp_drawable_offsets (old_layer->drawable_id, &x1, &y1);
  x1 -= (new_width - (int)old_layer->width) / 2;
  y1 -= (new_height - (int)old_layer->height) / 2;
  /*printf("Setting offset to (%d, %d)\n", x1, y1);*/
  gimp_layer_set_offsets (new_layer_id, x1, y1);

  g_free(new_name);
  if (dest)
    g_free(dest);

#ifdef GIMP_1_2
  gimp_undo_push_group_end (image_id);
#else /* GIMP_1_2 */
  gimp_image_undo_group_end (image_id);
#endif /* GIMP_1_2 */

  return arc_layer->drawable_id;
}

static gint
arclayer_dialog (void)
{
  GtkWidget *dlg;
  GtkWidget *toggle;
  GtkWidget *hbox;
  GtkWidget *logo_box;
  GtkWidget *frame;
  GtkWidget *table;
  GtkObject *scale_data;
#ifndef GIMP_1_2
  GdkPixbuf* icon_pixbuf;
  GtkWidget *preview;
#endif
  gimp_ui_init ("arclayer", TRUE);

#ifdef GIMP_1_2
  dlg = gimp_dialog_new (_("Arclayer"), "arclayer",
			 gimp_standard_help_func, "filters/arclayer.html",
			 GTK_WIN_POS_MOUSE,
			 FALSE, TRUE, FALSE,

			 GTK_STOCK_CANCEL, gtk_widget_destroy,
			 NULL, 1, NULL, FALSE, TRUE,

			 GTK_STOCK_OK, arclayer_ok_callback,
			 NULL, NULL, NULL, TRUE, FALSE,

			 NULL);
  g_signal_connect (G_OBJECT (dlg), "destroy",
                    G_CALLBACK (gtk_main_quit),
                    NULL);

#else /* GIMP 2.0 */
  dlg = gimp_dialog_new (_("Arclayer"), "arclayer",
                         NULL, 0,
                         gimp_standard_help_func, "filters/arclayer.html",

                         GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                         GTK_STOCK_OK,     GTK_RESPONSE_OK,

                         NULL);
  /*
  g_signal_connect (dlg, "response",
                    G_CALLBACK (arclayer_ok_callback),
                    &dlg);
  g_signal_connect (dlg, "destroy",
                    G_CALLBACK (gtk_widget_destroyed),
                    &dlg);
  */
#endif /* GIMP_1_2 */

  hbox = gtk_hbox_new (FALSE, 6);
  gtk_container_set_border_width (GTK_CONTAINER (hbox), 6);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dlg)->vbox), hbox, TRUE, TRUE, 0);

  logo_box = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), logo_box, FALSE, FALSE, 0);

  frame = gtk_frame_new (NULL);
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN);
  gtk_box_pack_start (GTK_BOX (logo_box), frame, FALSE, FALSE, 0);

#ifndef GIMP_1_2
  icon_pixbuf = gdk_pixbuf_new_from_inline(-1, arclayer_icon, FALSE, NULL);

  preview = gtk_image_new_from_pixbuf (icon_pixbuf);

  gtk_container_add (GTK_CONTAINER (frame), preview);
  gtk_widget_show (preview);
#endif /* GIMP_1_2 */
  gtk_widget_show (frame);
  gtk_widget_show (logo_box);

  frame = gtk_frame_new (_("Options"));
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_ETCHED_IN);
  gtk_box_pack_start (GTK_BOX (hbox), frame, FALSE, FALSE, 0);

  table = gtk_table_new (6, 3, FALSE);
  gtk_table_set_col_spacings (GTK_TABLE (table), 4);
  gtk_table_set_row_spacings (GTK_TABLE (table), 2);
  gtk_container_set_border_width (GTK_CONTAINER (table), 4);
  gtk_container_add (GTK_CONTAINER (frame), table);

  scale_data = gimp_scale_entry_new (GTK_TABLE (table), 0, 0,
				     _("_Radius:"), SCALE_WIDTH, 5,
				     avals.radius, 50.0, 1000.0, 1.0, 10.0, 1,
				     TRUE, 0, 0,
				     NULL, NULL);
  g_signal_connect (G_OBJECT (scale_data), "value_changed",
                    G_CALLBACK (gimp_int_adjustment_update),
                    &avals.radius);

  toggle = gtk_check_button_new_with_mnemonic (_("_Top"));
  gtk_table_attach(GTK_TABLE (table), toggle, 0, 1, 1, 2, 0, 0, 0, 0);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle), avals.top);
  gtk_widget_show (toggle);
  g_signal_connect (G_OBJECT (toggle), "toggled",
                    G_CALLBACK (gimp_toggle_button_update),
                    &avals.top);

  gtk_widget_show (table);
  gtk_widget_show (frame);
  gtk_widget_show (hbox);
  gtk_widget_show (dlg);

#ifdef GIMP_1_2
  gtk_main ();
  gdk_flush ();

#else /* ! GIMP_1_2 */
  /* Otherwise, code for gimp 2.0 */
  if (gimp_dialog_run (GIMP_DIALOG (dlg)) == GTK_RESPONSE_OK)
  {
    arclayer_ok_callback (dlg, GTK_RESPONSE_OK, 0);
    aint.run = TRUE;
  }

  gtk_widget_destroy (dlg);

#endif /* GIMP_1_2 */

  return aint.run;
}

static void
arclayer_ok_callback (GtkWidget *widget,
#ifndef GIMP_1_2
                      gint response_id,
#endif
                      gpointer   data)
{
  aint.run = TRUE;
  gtk_widget_destroy (GTK_WIDGET (data));
}

